unit Assembly;
{$I Common.inc}
//************************************************************************
//                                                         *
// :                                                              *
// CollectMinDissect --     ()             *
// CollectMinCover   --     ()               *
// CollectGradCover  --                  *
//  . 53                                                   *
//************************************************************************

interface

uses Root, Items, SetList;

//      ()

function CollectMinDissect(aUniv: TSet;      //  
                           aBuf: TBuffer   //   
                           ):TCostSet;

//     ()
//  . 53

function CollectMinCover(aUniv: TSet;      //  
                         aBuf: TBuffer   //   
                        ):TCostSet;

//        

function CollectGradCover(aBase: TSet;      //  
                          aBuf: TBuffer   //   
                         ):TCostSet;

///       

function CalcDifficult(aSet: TSet): Extended;

//////////////////////////////////////////////////////////////////////////
implementation
uses SetUtils;

//////////////////////////////////////////////////////////////////////////
// TSetBlock
//      
//    ()   ()
//////////////////////////////////////////////////////////////////////////

type

  TSetBlock = class (TSetList)
    mLabel : TItem;    //  
    constructor Create(aLabel: TItem);
    function Compare(arg: TItem): TCompare; override;
    procedure Print(var aFile: Text); override;
  end;

constructor TSetBlock.Create(aLabel: TItem);
begin
  inherited Create;
  mLabel:= aLabel;
end;

function TSetBlock.Compare(arg: TItem): TCompare;
begin
  Result:= cmpLess;  //      
end;

procedure TSetBlock.Print(var aFile: Text);
begin
  mLabel.Print(aFile);
  inherited;
end;

//////////////////////////////////////////////////////////////////////////
//       ,
//       ?
//////////////////////////////////////////////////////////////////////////

function TestCoverOrDissect(aUniv: TSet;     // 
                            aBuf: TBuffer  //  
                           ): boolean;
var  CS : TCostSet;  //    
    Sum : TSet;      //  
      i : integer;
begin
  Result:= false;
  Sum:= CreateSet;
  //      ,
  //     ,   
  for i:=1 to aBuf.GetCount do begin
    CS:= aBuf.GetByIndex(i) as TCostSet;
    Sum.Add(CS.mSet);
    if Sum.Compare(aUniv) in [cmpEq, cmpGreate] then begin
      Result:= True;
      Break;
    end;
  end;
  Sum.Free;
end;

//////////////////////////////////////////////////////////////////////////
///                          (DISSECTION)
//////////////////////////////////////////////////////////////////////////

//      

function GenFilter(aBuf: TBuffer): TBuffer;
var Ci, Cj : TCostSet;
      i, j : integer;
      Comp : TCompare;
begin
  //   :
  for i:= 1 to aBuf.GetCount do begin
    Ci:= aBuf.GetByIndex(i) as TCostSet;
    Ci.mFlag:= false;
  end;
  //     
  //        
  for i:= 1 to aBuf.GetCount-1 do begin
    Ci:= aBuf.GetByIndex(i) as TCostSet;
    if Ci.mFlag then Continue;             //  
    for j:= i+1 to aBuf.GetCount do begin
      Cj:= aBuf.GetByIndex(j) as TCostSet;
      if Cj.mFlag then Continue;           //  
      Comp:= Ci.mSet.Compare(Cj.mSet);     //  
      if Comp = cmpEq then begin
        //  ,  :
        if Ci.mCost >= Cj.mCost
          then Ci.mFlag:= true    // i-  
          else Cj.mFlag:= true;   // j-  
      end;
    end;
  end;

  Result:= TBuffer.Create;
  //        
  //     aBuf
  for i:= 1 to aBuf.GetCount do begin
    Ci:= aBuf.Get as TCostSet;
    aBuf.Put(Ci);
    //    ,     
    if not Ci.mFlag  then Result.Put(Ci);
  end;
end;

//////////////////////////////////////////////////////////////////////////
//      
// aSet  - 
// aBuf  -     ()
//////////////////////////////////////////////////////////////////////////

function GenDissectBlocks(aUniv: TSet; aBuf: TBuffer): TSet;
var i, j: integer;
    t: TItem;         //    
    S: TCostSet;      //   
    Block: TSetBlock; //  ( )
begin
  Result:= CreateSet;                  //  -
  for i:=1 to aUniv.GetCount do begin  //    
    if aBuf.GetCount=0 then Break;     //    
    //    :
    t:= aUniv.GetItem(i);
    Block:= TSetBlock.Create(t); //  
    //   :
    for j:= 1 to aBuf.GetCount do begin
      S:= aBuf.Get as TCostSet; //   
      if S.mSet.Exist(t)        //     
        then Block.Insert(S)    //    S  
        else aBuf.Put(S);       //      
    end;
    if Block.GetCount>0         //    
      then Result.Insert(Block) //    
      else Block.Free;          //    
  end;
  aBuf.Free;  //  
end;

//////////////////////////////////////////////////////////////////////////
//       
//////////////////////////////////////////////////////////////////////////

function CalcDifficult(aSet: TSet): Extended;
var S: TSet;  //  
begin
  Result:=1;
  //     ( )
  S:= aSet.GetFirst as TSet;
  while Assigned(S) do begin
    if S.GetCount=1
      then Result:= Result +1
      else Result:= Result * S.GetCount;
    S:= aSet.GetNext as TSet;
  end;
end;

//////////////////////////////////////////////////////////////////////////
//     
//  . 53
//////////////////////////////////////////////////////////////////////////

function CollectMinDissect(aUniv: TSet;      //  
                           aBuf: TBuffer   //   
                           ): TCostSet;
var
    Buf     : TBuffer;  //    
    Univ    : TSet;       //  
    BestCost: integer;    //  
    Blocks  : TSet;       //  
    Res : TCostSet;       //   
  //------------------------------------------------------------
  //    

  procedure BlockHandle(aBlock: integer;  //  
                        aCost: integer;   //  
                        aSum: TSet;       //  
                        aRes: TCostSet    //  
                        );
  var Block: TSetBlock; //  
      Sum : TSet;       //   
      Res : TCostSet;   //   ( )
       CS : TCostSet;   //     
  begin
    Block:= Blocks.GetItem(aBlock) as TSetBlock;
    //       
    if aSum.Exist(Block.mLabel) then begin
      //         (  )
      if aBlock < Blocks.GetCount then begin
        BlockHandle(aBlock+1, aCost, aSum, aRes) //    
      end
    end else begin
      //    
      CS:= Block.GetFirst as TCostSet;
      while Assigned(CS) do begin
        //      
        //    
        //  CS    . ,  
        if ((aCost+CS.mCost) < BestCost) and
           not aSum.TestIntersect(CS.mSet) then begin
          //     :
          Sum:= aSum.Copy as TSet;     //   
          Res:= aRes.Copy as TCostSet; //   
          Sum.Add(CS.mSet);     //  
          Res.Insert(CS);       //     
          // ,    
          if Sum.GetCount = aUniv.GetCount then begin
            //    .
            //    ,
            //    
            BestCost:= aCost + CS.mCost;   //   
            Result.Free;                   //   
            Result:= Res.Copy as TCostSet; //    
          end else begin
            //    ,    
            if aBlock < Blocks.GetCount
              then BlockHandle(aBlock+1, aCost + CS.mCost, Sum, Res)
          end; // else
          //       :
          //     
          Res.Free;   //  
          Sum.Free;   //  
        end; // if
        CS:= Block.GetNext as TCostSet;
      end; // while
    end; // else
  end;
  //------------------------------------------------------------
begin  { CollectMinDissect }

  Result:= TCostSet.CreateEmpty;  //   -

  //    ,  

  if not TestCoverOrDissect(aUniv, aBuf) then Exit;

  Buf:= GenFilter(aBuf);  //     

  Blocks:= GenDissectBlocks(aUniv, Buf);  //   
  //  :
  BestCost:= MaxInt;          //  
  Univ := CreateSet;          //  
  Res:= TCostSet.CreateEmpty; //   -

  //   :
  BlockHandle(1, 0, Univ, Res);

  // :
  Res.Free;                   //   
  Univ.Free;                  //  
  Blocks.ClrAndDestroy;       //  
  Blocks.Free;                //   
end;

//////////////////////////////////////////////////////////////////////////
//                          (COVER)
//////////////////////////////////////////////////////////////////////////

//       
// aUniv - 
// aBuf  -     ()

function GenCoverBlocks(aUniv: TSet; aBuf: TBuffer): TSet;
var i, j: integer;
    t: TItem;         //    aSet
    S: TCostSet;      //   
    Block: TSetBlock; //  ( )
begin
  Result:= CreateSet; //   
  //    :
  for i:=1 to aUniv.GetCount do begin
    t:= aUniv.GetItem(i);        //    
    Block:= TSetBlock.Create(t); //  
    //   :
    for j:= 1 to aBuf.GetCount do begin
      S:= aBuf.GetByIndex(j) as TCostSet; //   
      if S.mSet.Exist(t)      //     
        then Block.Insert(S)  //    S  
    end;
    Result.Insert(Block) //     
  end;
end;

//////////////////////////////////////////////////////////////////////////
//     ()
//  . 53
//////////////////////////////////////////////////////////////////////////

function CollectMinCover(aUniv: TSet;      //  
                         aBuf: TBuffer   //    
                        ): TCostSet;
var Buf: TBuffer;       //   
    Univ    : TSet;       //  
    BestCost: integer;    //  
    Blocks  : TSet;       //  
    Res : TCostSet;       //   
  //------------------------------------------------------------
  //    

  procedure BlockHandle(aBlock, aCost: integer; aSum: TSet; aRes: TCostSet);
  var Block : TSetBlock; //  
      Sum : TSet;        //  
      Res : TCostSet;    //   ( )
       CS : TCostSet;    //     
  begin
    //     
    Block:= Blocks.GetItem(aBlock) as TSetBlock;
    //        
    if aSum.Exist(Block.mLabel) then begin
      //   ,   
      //    ,    
      if aBlock<Blocks.GetCount then BlockHandle(aBlock+1,aCost,aSum,aRes)
    end else begin
      //    
      CS:= Block.GetFirst as TCostSet; //   
      //   
      while Assigned(CS) do begin
        //      
        if (aCost+CS.mCost) < BestCost then begin //   ?
          Sum:= aSum.Copy as TSet;     //   
          Res:= aRes.Copy as TCostSet; //   
          Sum.Add(CS.mSet);     //  
          Res.Insert(CS);       //  
          // ,     ():
          if Sum.GetCount = aUniv.GetCount then begin
            //   ,
            //    
            BestCost:= aCost + CS.mCost;   //   
            Result.Free;                   //   
            Result:= Res.Copy as TCostSet; //    
          end else begin
            //    ,    
            if aBlock < Blocks.GetCount
              then BlockHandle(aBlock+1, aCost + CS.mCost, Sum, Res);
          end; // else
          //       :
          Res.Free;  //  
          Sum.Free;  //  
        end; // then
        CS:= Block.GetNext as TCostSet;  //   
      end; // while
    end; // else
  end;
  //------------------------------------------------------------

begin  { CollectMinCover }

  Result:= TCostSet.CreateEmpty; //   

  // ,   ,  :
  if not TestCoverOrDissect(aUniv, aBuf) then Exit;

  Buf:= GenFilter(aBuf); //     
  Blocks:= GenCoverBlocks(aUniv, Buf);  //   
  Buf.Free;                            //    , 
  Res:= TCostSet.CreateEmpty; //   
  Univ := CreateSet;         //   ()
  BestCost:= MaxInt;         //  
  BlockHandle(1, 0, Univ, Res);  //   
  // :
  Res.Free;             //   
  Univ.Free;            //    ()
  Blocks.ClrAndDestroy; //  
  Blocks.Free;
end;

//////////////////////////////////////////////////////////////////////////
//       
//////////////////////////////////////////////////////////////////////////

function CollectGradCover(aBase: TSet;
                          aBuf: TBuffer
                         ): TCostSet;
var    Buf : TBuffer;    //   
      Univ : TSet;       //  
    Blocks : TSet;       // 
     Block : TSetBlock;  //  
    BestCS : TCostSet;   //   
         i : Integer;
  //------------------------------------------------------------
  function GetBest: TCostSet;
  var      CS : TCostSet;   //    
            S : TSet;       //  
    BestDelta : integer;    //  
     BestCost : integer;    //  
            i : integer;
  begin
    Result:= nil;
    S:= CreateSet;  //   
    BestDelta:=0;   //  
    BestCost:= 1;   //  
    //   :
    for i:= 1 to Block.GetCount do begin
      CS:= Block.GetItem(i) as TCostSet;  //  .  
      S.CopyItems(CS.mSet);               //  
      S.Sub(Univ);                        //   
      //    ,  
      if S.GetCount * BestCost > CS.mCost * BestDelta then begin
        Result:= CS;             //  
        BestCost:= CS.mCost;     //  
        BestDelta:= S.GetCount;  //  
      end;
    end;
    S.Free;  //   
  end;
  //------------------------------------------------------------

begin  { CollectGradCover }

  Result:= TCostSet.CreateEmpty;  //  =  

  //   ,   

  if not TestCoverOrDissect(aBase, aBuf) then Exit;

  Buf:= GenFilter(aBuf);//    
  Blocks:= GenCoverBlocks(aBase, Buf);  //  
  Buf.Free;             //    
  Univ:= CreateSet;     //   
  //   :
  for i:= 1 to Blocks.GetCount do begin
    Block:= Blocks.GetItem(i) as TSetBlock;
    if Univ.Exist(Block.mLabel)  //      
      then Continue;             //   
    BestCS:= GetBest;            //     
    if Assigned(BestCS) then begin
      //      :
      Univ.Add(BestCS.mSet);     //  
      Result.Insert(BestCS);     //  
    end;
  end;
  //  :
  Univ.Free;              //   
  Blocks.ClrAndDestroy;   //   
  Blocks.Free;
end;
//////////////////////////////////////////////////////////////////////////

end.

