unit Items;
{$I Common.inc}
//************************************************************************
//                                                      *
//                                                                       *
//  :                                                     *
// TItemChar -- -                                           *
// TItemStr  -- -                                           *
// TItemNum  -- -                                            *
// TBuffer   --                                        *
//************************************************************************

interface

uses Root;

type

  // TItemChar -- -
  TItemChar = class (TItem)
    private
      mData: Char;
    public
      constructor Create(arg: Char);
      function Compare(arg: TItem): TCompare; override;
      function Copy: TItem; override;
      procedure Print(var aFile: TextFile); override;
      function GetData: Char;
  end;

  // TItemStr  -- -
  TItemStr = class(TItem)
    private
      mData: String;
    public
      constructor Create(arg: string);
      function Compare(arg: TItem): TCompare; override;
      function Copy: TItem; override;
      procedure Print(var aFile: TextFile); override;
      function GetData: string;
  end;

  // TItemNum  -- -
  TItemNum = class (TItem)
    private
      mData: integer;
    public
      constructor Create(arg: integer);
      function Compare(arg: TItem): TCompare; override;
      function Copy: TItem; override;
      procedure Print(var aFile: TextFile); override;
      function GetData: integer;
  end;

  //    
  //  

  PStackRec = ^TStackRec;
  TStackRec = record
    mItem : TItem;            //  
    mNext : PStackRec;        //   
  end;

  //    :   

  TBuffer = class (TItem)
    private
      mCount: integer;                //  
      mHead, mQue: PStackRec;         //      
      procedure Clr(aDestroy: boolean);
    public
      destructor Destroy;  override;
      function Copy: TItem; override; //  
      procedure Put(arg: TItem);      //   
      function Get: TItem;            //   
      procedure Push(arg: TItem);     //   
      function Pop: TItem;            //   
      procedure Reversion;            //  
      function GetByIndex(aIndex: integer): TItem; //   
      function GetCount: integer;     //    
      procedure Clear;                //    
      procedure ClrAndDestroy;        //    
      procedure Print(var aFile: TextFile); override; //   
  end;

//////////////////////////////////////////////////////////////////////////
implementation
//////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////
// TItemChar -- -
//////////////////////////////////////////////////////////////////////////

function TItemChar.Compare(arg: TItem): TCompare;
begin
  Result:= cmpEq;
  if Self=arg then Exit;  // 
  if not (arg is TItemChar) then begin
    Result:=cmpIncomp;    // 
    Exit;
  end;
  if (arg as TItemChar).mData = mData
    then Result:=cmpEq
    else if (arg as TItemChar).mData > mData
            then Result:= cmpLess
            else Result:= cmpGreate;
end;

constructor TItemChar.Create(arg: Char);
begin
  inherited Create;
  mData:= arg;
end;

function TItemChar.Copy: TItem;
begin
  Result:= TItemChar.Create(mData);
end;


procedure TItemChar.Print(var aFile: TextFile);
begin
  Write(aFile, mData:2)
end;

function TItemChar.GetData: Char;
begin
  Result:= mData
end;

//////////////////////////////////////////////////////////////////////////
// TItemStr  -- -
//////////////////////////////////////////////////////////////////////////

function TItemStr.Copy: TItem;
begin
  Result:= TItemStr.Create(mData);
end;

constructor TItemStr.Create(arg: string);
begin
  inherited Create;
  mData:= arg
end;

function TItemStr.GetData: string;
begin
  Result:= mData;
end;

function TItemStr.Compare(arg: TItem): TCompare;
begin
  Result:= cmpEq;
  if Self=arg then Exit;  // 
  if not (arg is TItemStr) then begin
    Result:=cmpIncomp;    // 
    Exit;
  end;
  if (arg as TItemStr).mData <> mData
    then if (arg as TItemStr).mData > mData
            then Result:= cmpLess
            else Result:= cmpGreate
end;

procedure TItemStr.Print(var aFile: TextFile);
begin
  Write{ln}(aFile, ' ', mData);
end;

//////////////////////////////////////////////////////////////////////////
// TItemNum  -- -
//////////////////////////////////////////////////////////////////////////

function TItemNum.Compare(arg: TItem): TCompare;
begin
  Result:= cmpEq;
  if Self=arg then Exit;  // 
  if not (arg is TItemNum) then begin
    Result:=cmpIncomp;    // 
    Exit;
  end;
  if (arg as TItemNum).mData <> mData
    then if (arg as TItemNum).mData > mData
            then Result:= cmpLess
            else Result:= cmpGreate
end;


function TItemNum.Copy: TItem;
begin
  Result:= TItemNum.Create(mData);
end;

constructor TItemNum.Create(arg: integer);
begin
  inherited Create;
  mData:= arg
end;

function TItemNum.GetData: integer;
begin
  Result:= mData;
end;

procedure TItemNum.Print(var aFile: TextFile);
begin
  Write(aFile,' ', mData);
end;

//////////////////////////////////////////////////////////////////////////
// TBuffer --      
//////////////////////////////////////////////////////////////////////////

//     

function TBuffer.Copy: TItem;
var t: TItem;
    i: integer;
begin
  Result:= TBuffer.Create;
  for i:= 1 to GetCount do begin
    t:= Get;
    Put(t);
    TBuffer(Result).Put(t);
  end;
end;

//     

destructor TBuffer.Destroy;
begin
  Clear;
  inherited;
end;

//  

procedure TBuffer.Clr(aDestroy: boolean);
var i : integer;
    t : TItem;
begin
  for i:= 1 to mCount do begin
    t:= Get;
    if aDestroy then t.Free
  end;
end;

//    

procedure TBuffer.ClrAndDestroy;
begin
  Clr(true)
end;

//   

procedure TBuffer.Clear;
begin
  Clr(false)
end;

//   

function TBuffer.GetCount: integer;
begin
  Result:= mCount;
end;

// Push -    (  )

procedure TBuffer.Push(arg: TItem);
var r: PStackRec;
begin
  New(r);
  r^.mItem:= arg;
  r^.mNext:= mHead;
  mHead:= r;
  Inc(mCount);
  //     
  if not Assigned(mQue) then mQue:= r;
end;

// Put -    (  )

procedure TBuffer.Put(arg: TItem);
var r: PStackRec;
begin
  New(r);
  r^.mItem:= arg;
  r^.mNext:= nil;
  //    
  if Assigned(mQue) then mQue^.mNext:= r;
  mQue:= r;
  Inc(mCount);
  //    
  if not Assigned(mHead) then mHead:= r;
end;

//      (   )

function TBuffer.Get: TItem;
var r: PStackRec;
begin
  Result:= nil;
  if not Assigned(mHead) then Exit;
  r:= mHead;
  Result:= mHead^.mItem;
  mHead:= mHead^.mNext;
  Dispose(r);
  Dec(mCount);
  //    
  if mCount=0 then mQue:= nil;
end;

//    ( Get)

function TBuffer.Pop: TItem;
begin
  Result:= Get;
end;

//     

function TBuffer.GetByIndex(aIndex: integer): TItem;
var  r: PStackRec;
begin
  Result:= nil;
  if (aIndex<1) or (aIndex>mCount) then Exit;
  r:= mHead;
  while Assigned(r) do begin
    Dec(aIndex);
    if aIndex=0 then Break;
    r:= r^.mNext;
  end;
  if Assigned(r) then Result:= r^.mItem;
end;

//   (   )

procedure TBuffer.Reversion;
var B : TBuffer;
    t : TItem;
begin
  B:= TBuffer.Create;
  t:= Get;
  while Assigned(t) do begin   B.Push(t);  t:= Get;  end;
  t:= B.Get;
  while Assigned(t) do begin   Put(t);   t:= B.Get;  end;
  B.Free;
end;

//  

procedure TBuffer.Print(var aFile: TextFile);
var r: PStackRec;
begin
  Writeln(aFile, '(');
  r:= mHead;
  while Assigned(r) do begin
    r^.mItem.Print(aFile);
    r:= r^.mNext;
  end;
  Writeln(aFile, ') BufCount= ', mCount);
end;

//////////////////////////////////////////////////////////////////////////

end.
